package org.processmining.datapetrinets.dsl;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.petrinet.Marking;

import com.google.common.base.Joiner;

final class DPNBuildingVisitor implements DPNParserVisitor {

	public class NodePair {

		private final Place place;
		private final Transition transition;

		public NodePair(Place place, Transition transition) {
			this.place = place;
			this.transition = transition;
		}

		public Place getPlace() {
			return place;
		}

		public Transition getTransition() {
			return transition;
		}

	}

	public class PlaceToTransition extends NodePair {

		public PlaceToTransition(Place place, Transition transition) {
			super(place, transition);
		}

	}

	public class TransitionToPlace extends NodePair {

		public TransitionToPlace(Place place, Transition transition) {
			super(place, transition);
		}

	}

	private Map<String, PNWDTransition> transitionMap;
	private Map<String, Place> placeMap;
	private Map<String, DataElement> variableMap;

	private DataPetriNetsWithMarkings dpn;

	public Object visit(DPNVariables node, Object data) {
		node.childrenAccept(this, data);
		return null;
	}

	public Object visit(DPNVariable node, Object data) {
		String name = unquoteTransfromEscapes(node.name);
		DataElement variable = dpn.addVariable(name, convertType(node.type), null, null);
		variableMap.put(name, variable);
		return null;
	}

	private Class<?> convertType(Token type) {
		switch (type.kind) {
			case DPNParserConstants.LITERAL :
				return String.class;

			case DPNParserConstants.DISCRETE :
				return Long.class;

			case DPNParserConstants.BOOLEAN :
				return Boolean.class;

			case DPNParserConstants.CONTINUOUS :
				return Double.class;

			case DPNParserConstants.DATE :
				return Date.class;

			default :
				throw new IllegalStateException();
		}
	}

	public Object visit(DPNPlaceFlow node, Object data) {
		Place place = (Place) node.jjtGetChild(0).jjtAccept(this, null);
		Transition transition = (Transition) node.jjtGetChild(2).jjtAccept(this, null);
		node.jjtGetChild(1).jjtAccept(this, new PlaceToTransition(place, transition));
		return null;
	}

	public Object visit(DPNTransitionFlow node, Object data) {
		Place place = (Place) node.jjtGetChild(2).jjtAccept(this, null);
		Transition transition = (Transition) node.jjtGetChild(0).jjtAccept(this, null);
		node.jjtGetChild(1).jjtAccept(this, new TransitionToPlace(place, transition));
		return null;
	}

	public Object visit(DPNPlace node, Object data) {
		String name = unquoteTransfromEscapes(node.name.image);
		String uniqueName = genUniqueName(name, node.nodeId);
		Place place = placeMap.get(uniqueName);
		if (place == null) {
			place = dpn.addPlace(name);
			placeMap.put(uniqueName, place);
		}
		return place;
	}

	private static String genUniqueName(String name, Token nodeId) {
		return name + getUniqueId(nodeId);
	}

	private static String getUniqueId(Token nodeId) {
		return nodeId != null ? nodeId.image : "0";
	}

	public Object visit(DPNTransition node, Object data) {
		String name = unquoteTransfromEscapes(node.name.image);
		String uniqueName = genUniqueName(name, node.nodeId);
		PNWDTransition transition = transitionMap.get(uniqueName);
		if (transition == null) {
			transition = (PNWDTransition) dpn.addTransition(name);
			transitionMap.put(uniqueName, transition);
		}
		return transition;
	}

	public Object visit(DPNBiEdge node, Object data) {
		NodePair input = (NodePair) data;
		dpn.addArc(input.getPlace(), input.getTransition());
		dpn.addArc(input.getTransition(), input.getPlace());
		return null;
	}

	public Object visit(DPNEdge node, Object data) {
		NodePair input = (NodePair) data;
		if (input instanceof PlaceToTransition) {
			dpn.addArc(input.getPlace(), input.getTransition());
		} else {
			dpn.addArc(input.getTransition(), input.getPlace());
		}
		return null;
	}

	public Object visit(DPNGuard node, Object data) {
		String uniqueName = genUniqueName(unquoteTransfromEscapes(node.name.image), node.nodeId);
		PNWDTransition transition = transitionMap.get(uniqueName);
		if (transition != null) {
			try {
				String expressionInBrackets = extracted(node);
				String expressionPossiblyQuoted = expressionInBrackets.substring(1, expressionInBrackets.length() - 1);
				String expression = unquoteTransfromEscapes(expressionPossiblyQuoted);
				transition.setGuard(dpn, expression);
			} catch (NonExistingVariableException e) {
				throw new DPNBuildingError(node.expression, "Missing variables "
						+ Joiner.on(",").join(e.getMissingVariables()) + " in expression " + extracted(node)
						+ " for transition " + node.name, e);
			} catch (ParseException e) {
				DPNBuildingError buildingError = new DPNBuildingError(node.expression,
						"Invalid expression syntax for transition " + node.name, e);
				buildingError.setExpectedTokenSequence(e.expectedTokenSequences);
				buildingError.setTokenImage(e.tokenImage);
				throw buildingError;
			}
		} else {
			throw new DPNBuildingError(node.name, "Missing transition " + node.name + " cannot add expression "
					+ extracted(node) + "!");
		}
		return null;
	}

	private String extracted(DPNGuard node) {
		return node.expression.image;
	}

	public Object visit(DPNWrite node, Object data) {
		String uniqueName = genUniqueName(unquoteTransfromEscapes(node.name.image), node.nodeId);
		PNWDTransition transition = transitionMap.get(uniqueName);
		if (transition != null) {
			String[] splittedWrites = node.writes.image.substring(1, node.writes.image.length() - 1).split(",");
			for (String varName : splittedWrites) {
				DataElement variable = variableMap.get(varName);
				if (variable != null) {
					dpn.assignWriteOperation(transition, variable);
				} else {
					throw new DPNBuildingError(node.writes, "Missing variable declaration " + varName);
				}
			}
		} else {
			throw new DPNBuildingError(node.name, "Missing transition declaration " + node.name);
		}
		return null;
	}

	public Object visit(DPNInitial node, Object data) {
		String uniqueName = genUniqueName(unquoteTransfromEscapes(node.name.image), node.nodeId);
		Place place = placeMap.get(uniqueName);
		if (place != null) {			
			createInitialMarkingIfNeeded().add(place, node.tokens);
		} else {
			throw new DPNBuildingError(node.name, "Missing place " + node.name);
		}
		return null;
	}

	public Object visit(DPNFinal node, Object data) {
		String uniqueName = genUniqueName(unquoteTransfromEscapes(node.name.image), node.nodeId);
		Place place = placeMap.get(uniqueName);
		if (place != null) {
			createFinalMarkingIfNeeded().add(place, node.tokens);
		} else {
			throw new DPNBuildingError(node.name, "Missing place " + node.name);
		}
		return null;
	}
	
	private Marking createInitialMarkingIfNeeded() {
		if (dpn.getInitialMarking() == null) {
			dpn.setInitialMarking(new Marking());
		}
		return dpn.getInitialMarking();
	}
	
	private Marking createFinalMarkingIfNeeded() {
		if (dpn.getFinalMarkings() == null) {
			dpn.setFinalMarkings(new Marking[] { new Marking() });
		}
		return dpn.getFinalMarkings()[0];
	}

	public Object visit(DPNInvisible node, Object data) {
		String uniqueName = genUniqueName(unquoteTransfromEscapes(node.name.image), node.nodeId);
		Transition transition = transitionMap.get(uniqueName);
		if (transition != null) {
			transition.setInvisible(true);
		} else {
			throw new DPNBuildingError(node.name, "Missing transition " + node.name);
		}
		return null;
	}

	public Object visit(DPNRoot node, Object data) {
		transitionMap = new HashMap<>();
		placeMap = new HashMap<>();
		variableMap = new HashMap<>();
		if (node.jjtGetValue() == null) {
			dpn = new PetriNetWithData("DPN");
		} else {
			dpn = new PetriNetWithData(unquoteTransfromEscapes((String) node.jjtGetValue()));
		}
		node.childrenAccept(this, data);
		return null;
	}

	public Object visit(SimpleNode node, Object data) {
		throw new IllegalStateException();
	}

	public DataPetriNetsWithMarkings getDPN() {
		return dpn;
	}

	private static String unquoteTransfromEscapes(String label) {
		if (label.charAt(0) == '\"') {
			String unquoted = label.substring(1, label.length() - 1);
			unquoted = unquoted.replace("\\\"", "\"");
			return unquoted;
		} else {
			return label;
		}
	}

}