package org.processmining.datapetrinets.dsl.editor;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.dsl.DPN;
import org.processmining.datapetrinets.dsl.editor.DPNTextEditor.ModelSaver;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.utils.ConnectionManagerHelper;
import org.processmining.plugins.utils.ProvidedObjectHelper;

public final class PNTextEditorPlugin {

	private static final String TEMPLATE = "[NAME]\r\n" + 
			"\"my dpn\"\r\n" + 
			"\r\n" + 
			"[PLACES]\r\n" + 
			"p1 -> A\r\n" + 
			"p2 -> B\r\n" + 
			"p2 -> C\r\n" + 
			"p2 -> skip\r\n" + 
			"\r\n" + 
			"[TRANSITIONS]\r\n" + 
			"A -> p2\r\n" + 
			"B -> p3\r\n" + 
			"C -> p3\r\n" + 
			"skip -> p3\r\n" + 
			"\r\n" + 
			"[VARIABLES]\r\n" + 
			"X [LITERAL]\r\n" + 
			"\r\n" + 
			"[WRITES]\r\n" + 
			"A [X]\r\n" + 
			"\r\n" + 
			"[GUARDS]\r\n" + 
			"B [\"X==\\\"accept\\\"\"]\r\n" + 
			"C [\"X==\\\"reject\\\"\"]\r\n" + 
			"\r\n" + 
			"[INITIAL]\r\n" + 
			"p1 [1]\r\n" + 
			"\r\n" + 
			"[FINAL]\r\n" + 
			"p3 [1]\r\n" + 
			"\r\n" + 
			"[INVISIBLE]\r\n" + 
			"skip";

	@Plugin(name = "Create DPN (Text language based)", parameterLabels = {}, returnLabels = { "Petri Net with Data" }, returnTypes = { DPNTextEditor.class }, //
			userAccessible = true, level = PluginLevel.Regular, help = "Create a DPN (Petri net with data) by using a domain specific language.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public static DPNTextEditor createDPN(PluginContext context) {
		return createDPNEditor(context.getParentContext(), TEMPLATE);
	}

	@Plugin(name = "Edit DPN (Text language based)", parameterLabels = { "Petri net" }, returnLabels = { "Petri Net with Data" }, returnTypes = { DPNTextEditor.class }, //
			userAccessible = true, level = PluginLevel.Regular,  help = "Edits a DPN (Petri net with data) by using a domain specific language.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public static DPNTextEditor editDPN(PluginContext context, PetrinetGraph net) {
		DataPetriNet dpn = convertToDPN(net, context);
		Marking initialMarking = getInitialMarking(context, dpn);
		Marking finalMarking = getFinalMarking(context, dpn);
		return createDPNEditor(context.getParentContext(), DPN.toText(dpn, initialMarking, finalMarking));
	}

	private static DPNTextEditor createDPNEditor(final PluginContext context, String initialCode) {
		return DPNTextEditor.newInstance(initialCode, new ModelSaver() {
			
			public void save(DataPetriNetsWithMarkings dpnWithMarkings) {
				String netLabel = dpnWithMarkings.getLabel();			
				ProvidedObjectHelper.publish(context, netLabel, dpnWithMarkings, DataPetriNet.class, false);
				ProvidedObjectHelper.setFavorite(context, dpnWithMarkings);
				if (dpnWithMarkings.getInitialMarking() != null) {
					ProvidedObjectHelper.publish(context, "Initial marking for "+netLabel, dpnWithMarkings.getInitialMarking(), Marking.class, false);
					context.getConnectionManager().addConnection(new InitialMarkingConnection(dpnWithMarkings, dpnWithMarkings.getInitialMarking()));
				}				
				if (dpnWithMarkings.getFinalMarkings() != null) {
					ProvidedObjectHelper.publish(context, "Final marking for "+netLabel, dpnWithMarkings.getFinalMarkings()[0], Marking.class, false);
					context.getConnectionManager().addConnection(new FinalMarkingConnection(dpnWithMarkings, dpnWithMarkings.getFinalMarkings()[0]));
				}								
			}
		});
	}

	private static Marking getFinalMarking(PluginContext context, PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking[] finalMarkings = ((DataPetriNetsWithMarkings) net).getFinalMarkings();
			if (finalMarkings != null) {
				return finalMarkings[0];
			} else {
				return new Marking();
			}			
		} else {
			try {
				FinalMarkingConnection finalMarkingConnection = ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), FinalMarkingConnection.class, net);
				Marking finalMarking = finalMarkingConnection.getObjectWithRole(FinalMarkingConnection.MARKING);
				return finalMarking;
			} catch (ConnectionCannotBeObtained e) {
				return new Marking();
			}
		}
	}

	private static Marking getInitialMarking(PluginContext context, PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking initialMarking = ((DataPetriNetsWithMarkings) net).getInitialMarking();
			if (initialMarking != null) {
				return initialMarking;
			} else {
				return new Marking();
			}			
		} else {
			try {
				InitialMarkingConnection initialMarkingConnection = ConnectionManagerHelper
						.safeGetFirstConnection(context.getConnectionManager(), InitialMarkingConnection.class, net);
				Marking initialMarking = initialMarkingConnection.getObjectWithRole(InitialMarkingConnection.MARKING);
				return initialMarking;
			} catch (ConnectionCannotBeObtained e) {
				return new Marking();
			}
		}
	}

	private static DataPetriNet convertToDPN(PetrinetGraph net, PluginContext context) {
		if (net instanceof DataPetriNet) {
			return (DataPetriNet) net;
		} else {
			PetriNetWithDataFactory dpnFactory = new PetriNetWithDataFactory(net, net.getLabel(), true);
			dpnFactory.cloneInitialAndFinalConnection(context);
			return dpnFactory.getRetValue();
		}
	}

}
