package org.processmining.datapetrinets.io;

import java.awt.Dimension;
import java.awt.Point;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;

import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.processmining.contexts.uitopia.annotations.UIImportPlugin;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.datapetrinets.utils.MarkingsHelper;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.pnml.base.Pnml;
import org.processmining.plugins.utils.ProvidedObjectHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xmlpull.v1.XmlPullParserException;

import com.google.common.collect.ImmutableSet;
import com.google.common.io.ByteSource;
import com.google.common.io.ByteStreams;

@Plugin(name = "Import Data Petri Net from PNML file", level = PluginLevel.PeerReviewed, parameterLabels = { "Filename" }, returnLabels = { "Petri Net with Data" },//
returnTypes = { PetriNetWithData.class })
@UIImportPlugin(description = "PNML Petri Net with Data files", extensions = { "pnml" })
public class DataPetriNetImporter extends AbstractImportPlugin {

	public interface DPNWithLayout {
		
		DataPetriNetsWithMarkings getDPN();

		GraphLayoutConnection getLayout();
		
	}

	protected FileFilter getFileFilter() {
		return new FileNameExtensionFilter("PNML files", "pnml");
	}

	public DPNWithLayout importFromStream(InputStream inputStream) throws DPNIOException {

		try {
			byte[] dpnInput = ByteStreams.toByteArray(inputStream);
			ByteSource source = ByteSource.wrap(dpnInput);

			Pnml pnml = readBasicPnml(source);

			final PetriNetWithData petriWithData = new PetriNetWithData(pnml.getLabel());
			Marking initialMarking = new Marking();
			Marking finalMarking = new Marking();
			final GraphLayoutConnection layout = new GraphLayoutConnection(petriWithData);

			pnml.convertToNet(petriWithData, initialMarking, layout);

			if (!initialMarking.isEmpty()) {
				petriWithData.setInitialMarking(initialMarking);
			} else {
				petriWithData.setInitialMarking(null);
			}

			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			try (InputStream stream = source.openStream()) {
				Document document = factory.newDocumentBuilder().parse(new InputSource(stream));
				addFinalMarking(petriWithData, finalMarking, document);
				addVariablesToNet(petriWithData, document, layout);
				addOperationsToTransition(petriWithData, document);
			}

			return new DPNWithLayout() {

				public GraphLayoutConnection getLayout() {
					return layout;
				}

				public DataPetriNetsWithMarkings getDPN() {
					return petriWithData;
				}
			};
		} catch (InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException
				| NoSuchMethodException | SecurityException | ClassNotFoundException | IOException
				| XmlPullParserException | SAXException | ParserConfigurationException | NonExistingVariableException
				| ParseException e) {
			throw new DPNIOException(e);
		}
	}

	private Pnml readBasicPnml(ByteSource source) throws XmlPullParserException, IOException {
		try (InputStream stream = source.openStream()) {
			return PnmlImport.importPnml(stream);
		}
	}

	@Override
	protected Object importFromStream(PluginContext context, InputStream input, String filename, long fileSizeInBytes)
			throws Exception {
		if (fileSizeInBytes == 0) {
			throw new IllegalArgumentException("Trying to import an empty file");
		}

		DPNWithLayout dpnWithLayout = importFromStream(input);
		context.addConnection(dpnWithLayout.getLayout());

		DataPetriNetsWithMarkings dpnWithMarkings = dpnWithLayout.getDPN();

		if (MarkingsHelper.hasNonEmptyInitialMarking(context, dpnWithMarkings)) {
			ProvidedObjectHelper.publish(context, "Initial marking of " + dpnWithMarkings.getLabel(),
					dpnWithMarkings.getInitialMarking(), Marking.class, false);
			context.addConnection(new InitialMarkingConnection(dpnWithMarkings, dpnWithMarkings.getInitialMarking()));
		}

		if (MarkingsHelper.hasNonEmptyFinalMarking(context, dpnWithMarkings)) {
			ProvidedObjectHelper.publish(context, "Final marking of " + dpnWithMarkings.getLabel(),
					dpnWithMarkings.getFinalMarkings()[0], Marking.class, false);
			context.addConnection(new FinalMarkingConnection(dpnWithMarkings, dpnWithMarkings.getFinalMarkings()[0]));
		}

		context.getFutureResult(0).setLabel(dpnWithMarkings.getLabel());
		return dpnWithMarkings;
	}

	private void addFinalMarking(PetriNetWithData petriWithData, Marking finalMarking, Document document) {
		NodeList placeNodes = document.getElementsByTagName(PnmlHelper.PLACE);
		for (int i = 0; i < placeNodes.getLength(); i++) {
			Element placeNode = (Element) placeNodes.item(i);
			Place place = getPlace(petriWithData, placeNode);
			if (place != null) {
				NodeList nodeList = placeNode.getElementsByTagName(PnmlHelper.FINAL_MARKING_NODE);
				if (nodeList.getLength() > 0) {
					Element finalMarkingNode = (Element) nodeList.item(0);
					NodeList textNodes = finalMarkingNode.getElementsByTagName(PnmlHelper.TEXT);
					if (textNodes.getLength() == 1) {
						Node tokensNode = textNodes.item(0);
						String textContent = tokensNode.getTextContent();
						try {
							int numTokens = Integer.parseInt(textContent);
							finalMarking.add(place, numTokens);
						} catch (NumberFormatException e) {
							// We are lenient
							System.err.println("Wrong format for final marking of place " + place.getLabel() + " "
									+ textContent);
						}
					}
				}
			} else {
				// Does not support final markings as no localId is present
			}
		}

		if (!finalMarking.isEmpty()) {
			petriWithData.setFinalMarkings(new Marking[] { finalMarking });
		} else {
			petriWithData.setFinalMarkings(null);
		}
	}

	private Place getPlace(PetriNetWithData petriWithData, Element placeNode) {
		String localId = PnmlHelper.getLocalId(placeNode);
		if (localId != null) {
			for (Place p : petriWithData.getPlaces()) {
				if (localId.equals(p.getLocalID().toString())) {
					return p;
				}
			}
		} else {
			// Fallback and try with label
			String placeLabel = PnmlHelper.getLabel(placeNode);
			if (placeLabel != null) {
				for (Place p : petriWithData.getPlaces()) {
					if (placeLabel.equals(p.getLabel())) {
						return p;
					}
				}
			} else {
				return null;
			}

		}
		return null;
	}

	private void addOperationsToTransition(PetriNetWithData net, Document document)
			throws NonExistingVariableException, ParseException {
		NodeList nList = document.getElementsByTagName(PnmlHelper.TRANSITION);
		for (int i = 0; i < nList.getLength(); i++) {
			Element transitionNode = (Element) nList.item(i);
			String localId = PnmlHelper.getLocalId(transitionNode);
			String label = PnmlHelper.getLabel(transitionNode);
			for (Transition t : net.getTransitions()) {
				PNWDTransition trans = (PNWDTransition) t;
				if ((localId != null && localId.equals(trans.getLocalID().toString())) || localId == null
						&& label != null && label.equals(trans.getLabel())) {
					String invisible = transitionNode.getAttribute(PnmlHelper.INVISIBLE);
					if (invisible != null && invisible.trim().toLowerCase().equals(PnmlHelper.TRUE))
						trans.setInvisible(true);
					String guard = transitionNode.getAttribute(PnmlHelper.GUARD);
					if (guard != null && !guard.equals("")) {
						trans.setGuard(net, GuardExpression.Factory.newInstance(guard));
					}
					NodeList nList2 = transitionNode.getElementsByTagName(PnmlHelper.WRITE_VARIABLE);
					for (int j = 0; j < nList2.getLength(); j++) {
						String varName = nList2.item(j).getTextContent();
						try {
							net.assignWriteOperation(trans, net.getVariable(varName));
						} catch (IllegalArgumentException e) {
							throw new NonExistingVariableException(ImmutableSet.of(varName));
						}
					}
					nList2 = transitionNode.getElementsByTagName(PnmlHelper.READ_VARIABLE);
					for (int j = 0; j < nList2.getLength(); j++) {
						String varName = nList2.item(j).getTextContent();
						net.assignReadOperation(trans, net.getVariable(varName));
					}
				}
			}
		}
	}

	private void addVariablesToNet(PetriNetWithData net, Document doc, GraphLayoutConnection layout)
			throws InstantiationException, IllegalAccessException, IllegalArgumentException, InvocationTargetException,
			NoSuchMethodException, SecurityException, ClassNotFoundException {
		NodeList nList = doc.getElementsByTagName(PnmlHelper.VARIABLES);
		if (nList.getLength() > 0) {
			Element variablesElement = (Element) nList.item(0);
			nList = variablesElement.getElementsByTagName(PnmlHelper.VARIABLE);
			for (int i = 0; i < nList.getLength(); i++) {
				Element varNode = (Element) nList.item(i);
				Class<?> varType = Class.forName(varNode.getAttribute(PnmlHelper.TYPE));
				String varName = varNode.getElementsByTagName(PnmlHelper.NAME).item(0).getFirstChild().getTextContent();

				Comparable<?> minValue = null, maxValue = null;
				String minValueAttribute = varNode.getAttribute(PnmlHelper.MIN_VALUE);
				if (minValueAttribute != null && !minValueAttribute.equals(""))
					minValue = (Comparable<?>) varType.getConstructor(String.class).newInstance(minValueAttribute);
				String maxValueAttribute = varNode.getAttribute(PnmlHelper.MAX_VALUE);
				if (maxValueAttribute != null && !maxValueAttribute.equals(""))
					maxValue = (Comparable<?>) varType.getConstructor(String.class).newInstance(maxValueAttribute);
				DataElement dataElem = net.addVariable(varName, varType, minValue, maxValue);

				Element positionElement = (Element) varNode.getElementsByTagName(PnmlHelper.POSITION).item(0);
				if (positionElement != null) {
					try {
						int xPos = Integer.parseInt(positionElement.getAttribute(PnmlHelper.X));
						int yPos = Integer.parseInt(positionElement.getAttribute(PnmlHelper.Y));
						if (layout != null)
							layout.setPosition(dataElem, new Point(xPos, yPos));
					} catch (NumberFormatException e) {
					}
				}
				Element dimensionElement = (Element) varNode.getElementsByTagName(PnmlHelper.DIMENSION).item(0);
				if (dimensionElement != null) {
					try {
						int height = Integer.parseInt(dimensionElement.getAttribute(PnmlHelper.HEIGHT));
						int width = Integer.parseInt(dimensionElement.getAttribute(PnmlHelper.WIDTH));
						if (layout != null)
							layout.setSize(dataElem, new Dimension(width, height));
					} catch (NumberFormatException e) {
						// Ignore as we do not necessarily require the layout
					}
				}
				net.graphElementChanged(dimensionElement);
			}
		}
	}

}
