package org.processmining.datapetrinets.utils;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter.DecorationKey;
import org.processmining.models.graphbased.AbstractGraphElement;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

import com.google.common.escape.CharEscaperBuilder;
import com.google.common.escape.Escaper;

public class TikZUtil {

	private static final Escaper LATEX_ESCAPER = new CharEscaperBuilder().addEscape('#', "\\#").addEscape('$', "\\$")
			.addEscape('%', "\\%").addEscape('&', "\\&").addEscape('\\', "\textbackslash{}")
			.addEscape('^', "\textasciicircum{}").addEscape('_', "\\_").addEscape('{', "\\{").addEscape('}', "\\}")
			.addEscape('~', "\\textasciitilde{}}").toEscaper();

	private Map<AbstractGraphElement, Map<DecorationKey, Object>> decoration = new HashMap<>();

	public TikZUtil() {
		super();
	}

	public String transformToTikZ(PetrinetGraph net, Marking initialMarking, Marking finalMarking) {
		StringBuilder sb = new StringBuilder();

		sb.append("%%%%%%%% TikZ preamble %%%%%%%%%%%\n");
		sb.append(getPreamble() + "\n");
		sb.append("%%%%%%%% TikZ preamble %%%%%%%%%%%\n\n\n");

		sb.append("\\resizebox{\\textwidth}{!}{%\n");
		sb.append("\\begin{tikzpicture}%\n");
		Map<PetrinetNode, String> placeToId = calcHumanFriendlyIds(net.getPlaces(), "p");
		if (initialMarking.size() == 1) {
			String anchorNode = placeToId.get(initialMarking.iterator().next());
			sb.append("\\begin{scope}[petrinet,layered layout,anchor node=" + anchorNode
					+ ",grow=right,level distance=1cm,sibling distance=1cm]%\n");
		} else {
			sb.append("\\begin{scope}[petrinet,layered layout,grow=right,level distance=1cm,sibling distance=1cm]%\\n");
		}
		Map<PetrinetNode, String> transitionToId = calcHumanFriendlyIds(net.getTransitions(), "t");
		for (Place place : net.getPlaces()) {
			String placeLabel = calcTexFriendlyLabel(escapeLatex(place.getLabel()));
			Integer tokens = initialMarking.occurrences(place);
			if (tokens > 0) {
				sb.append(String.format(
						"\\node [place,tokens=%1d](%2$s) [label={[caption,font=\\tiny]below:%3$s}] {}; %%: %3$s \n",
						tokens, placeToId.get(place), placeLabel));
			} else if (finalMarking.contains(place)) {
				sb.append(String.format(
						"\\node [final-place](%2$s) [label={[caption,font=\\tiny]below:%3$s}] {}; %%: %3$s \n", tokens,
						placeToId.get(place), placeLabel));
			} else {
				sb.append(
						String.format("\\node [place](%1$s) [label={[caption,font=\\tiny]below:%2$s}] {}; %%: %2$s \n",
								placeToId.get(place), placeLabel));
			}
		}
		for (Transition transition : net.getTransitions()) {
			String transitionLabel = calcTexFriendlyLabel(escapeLatex(transition.getLabel()));
			if (transition.isInvisible()) {
				sb.append(String.format(
						"\\node [tau-transition](%1$s) [label={[caption,font=\\tiny]below:%2$s}] {}; %%: %2$s \n",
						transitionToId.get(transition), transitionLabel));
			} else {
				sb.append(String.format("\\node [transition](%1$s) [label={[caption]below:%2$s}] {};\n",
						transitionToId.get(transition), transitionLabel));
			}
		}

		sb.append("\\draw \n");
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getEdges()) {
			if (edge instanceof Arc) {
				if (edge.getSource() instanceof Place) {
					sb.append(String.format("(%1$s) edge[%2$s] (%3$s)\n", placeToId.get(edge.getSource()), "->",
							transitionToId.get(edge.getTarget())));
				} else {
					sb.append(String.format("(%1$s) edge[%2$s] (%3$s)\n", transitionToId.get(edge.getSource()), "->",
							placeToId.get(edge.getTarget())));
				}
			}
		}
		sb.deleteCharAt(sb.length() - 1); // remove last linebreak
		sb.append(";\n");
		sb.append("\\end{scope}\n");
		sb.append("\\end{tikzpicture}");
		sb.append("}%");
		return sb.toString();
	}

	private static final Pattern PLACE_PATTERN = Pattern.compile("p([0-9]+)");
	private static final Pattern TRANSITION_PATTERN = Pattern.compile("(t|tau)([0-9]+)");
	
	private String calcTexFriendlyLabel(String label) {
		Matcher placeMatcher = PLACE_PATTERN.matcher(label);
		if (placeMatcher.matches()) {
			return "$p_{"+placeMatcher.group(1)+"}$";
		}
		Matcher transitionMatcher = TRANSITION_PATTERN.matcher(label);
		if (transitionMatcher.matches()) {
			return "$\\tau_{"+transitionMatcher.group(2)+"}$";
		}
		return label;
	}

	private String getPreamble() {
		String tikzpackage = "\\usepackage{tikz}\r\n"
				+ "\\usetikzlibrary{arrows,automata,petri,shapes,shadows,positioning,decorations,backgrounds,graphdrawing,graphs}\r\n"
				+ "\\usegdlibrary{trees,layered,force}\r\n";
		String tikzset = "\\tikzset{\r\n" + "	petrinet/.style={font=\\scriptsize,>=latex,			\r\n"
				+ "		every edge/.append style={rounded corners},\r\n"
				+ "		every token/.append style={minimum size=1mm},\r\n"
				+ "		place/.style={circle,draw=black,inner sep=0pt,minimum size=4mm},\r\n"
				+ "		emph-place/.style={place,draw=red},\r\n"
				+ "		transition/.style={rectangle,draw=black,minimum size=5mm},\r\n"
				+ "		tau/.style={transition,fill=black,inner sep=0pt},\r\n"
				+ "		tau-transition-vertical/.style={tau,minimum height=5mm,minimum width=1mm},\r\n"
				+ "		tau-transition-horizontal/.style={tau,minimum height=1mm,minimum width=5mm},\r\n"
				+ "		tau-transition/.style={tau-transition-vertical},\r\n"
				+ "		caption/.style={align=center, outer sep=0pt},		\r\n" + "	}\r\n" + "}";
		return tikzpackage + tikzset;
	}

	private String escapeLatex(String label) {
		return LATEX_ESCAPER.escape(label);
	}

	private Map<PetrinetNode, String> calcHumanFriendlyIds(Collection<? extends PetrinetNode> nodes, String prefix) {
		Map<PetrinetNode, String> ids = new HashMap<>();
		for (PetrinetNode node : nodes) {
			int i = 0;
			String label = node.getLabel();
			while (ids.containsKey(label)) {
				label = node.getLabel() + (++i);
			}
			ids.put(node, label);
		}
		return ids;
	}

	public void setDecoration(Map<AbstractGraphElement, Map<DecorationKey, Object>> decoration) {
		this.decoration = decoration;
	}

	public Map<AbstractGraphElement, Map<DecorationKey, Object>> getDecoration() {
		return decoration;
	}

}
