package org.processmining.datapetrinets.validation;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.SortedSet;

import javax.swing.JOptionPane;

import org.cpntools.accesscpn.engine.highlevel.CheckerException;
import org.cpntools.accesscpn.engine.highlevel.HighLevelSimulator;
import org.cpntools.accesscpn.engine.highlevel.LocalCheckFailed;
import org.cpntools.accesscpn.engine.highlevel.PageSorter;
import org.cpntools.accesscpn.engine.highlevel.checker.Checker;
import org.cpntools.accesscpn.engine.highlevel.checker.ErrorInitializingSMLInterface;
import org.cpntools.accesscpn.engine.highlevel.instance.Binding;
import org.cpntools.accesscpn.engine.highlevel.instance.Instance;
import org.cpntools.accesscpn.engine.highlevel.instance.Marking;
import org.cpntools.accesscpn.engine.highlevel.instance.State;
import org.cpntools.accesscpn.model.FusionGroup;
import org.cpntools.accesscpn.model.HLDeclaration;
import org.cpntools.accesscpn.model.Page;
import org.cpntools.accesscpn.model.PetriNet;
import org.cpntools.accesscpn.model.RefPlace;
import org.cpntools.accesscpn.model.Transition;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.plugins.cpnet.ColouredPetriNet;
import org.processmining.plugins.cpnet.Logger;
import org.processmining.plugins.utils.ProvidedObjectHelper;

@Plugin(name = " Analyze Colour Petri Nets for Soundness", level = PluginLevel.Regular, parameterLabels = {
"Colour Petri Net", "Data Petri Net" }, returnLabels = { "Analysis Result" }, returnTypes = {
		CPNAnalysisResults.class }, userAccessible = true, keywords = { "DPN",
"conversion", "soundness", "verification"}, help = "Perform the soundness analysis of a Data Petri Net on the basis of the corresponding Colour Petri Net")
public class SoundnessOfPetriNets {

	@PluginVariant(variantLabel = "CPN Petri net", requiredParameterLabels = { 0, 1 })
	@UITopiaVariant(affiliation = "TU/e", author = "Massimiliano de Leoni,", email = "m.d.leoni@tue.nl")
	public static CPNAnalysisResults analysisCPN(UIPluginContext context, PetriNet cpn, DataPetriNet dpn) throws Exception  {
		return analysisCPN(context, new ColouredPetriNet(cpn), dpn);
	}
	
	/*@Plugin(name = " Analyze Colour Petri Nets for Soundness", level = PluginLevel.Regular, parameterLabels = {
	"Colour Petri Net", "Data Petri Net" }, returnLabels = { "Analysis Result" }, returnTypes = {
			CPNAnalysisResults.class }, userAccessible = true, keywords = { "DPN", "CPN",
	"soundness", "verification" }, help = "Perform the soundness analysis of a Data Petri Net on the basis of the corresponding Colour Petri Net")*/
	@PluginVariant(variantLabel = "Colour Petri net", requiredParameterLabels = { 0, 1 })
	@UITopiaVariant(affiliation = "TU/e", author = "Massimiliano de Leoni,", email = "m.d.leoni@tue.nl, f.mannhardt@tue.nl")
	public static CPNAnalysisResults analysisCPN(final UIPluginContext context, ColouredPetriNet cpn, DataPetriNet dpn) throws Exception  {
		int numStates=0;
		LinkedList<State> list=new LinkedList<State>();
		HashSet<org.cpntools.accesscpn.model.Transition> notObservedTransition=
				new HashSet<org.cpntools.accesscpn.model.Transition>();
		HashMap<State,Binding[]> pathTo=new HashMap<State,Binding[]>();
		LinkedList<State> deadlocks=new LinkedList<State>();


		
		HighLevelSimulator sim = checkModel(cpn.getPetriNet(), context.getProgress(), new Logger() {
			@Override
			public void log(final String message) {
				context.log(message);
			}
		});
		sim.initialState();
		
		State marking=sim.getMarking();
		pathTo.put(marking, new Binding[0]);
		list.add(marking);
		
		context.log("Computing the reachability graph. It may take a while...");
		
		
		for(Instance<Transition> trans : sim.getAllTransitionInstances())
			notObservedTransition.add(trans.getNode());

		List<Instance<Transition>> allTransitions = sim.getAllTransitionInstances();
		while(!list.isEmpty())
		{
			State oldMarking = marking;
			marking = list.removeFirst();
			sim.setMarkingFast(marking, oldMarking, true);
			//setMarking(marking);
			boolean enabled=false;
			for(Instance<? extends Transition> trans : sim.isEnabled(allTransitions))
			{

				enabled=true;
				notObservedTransition.remove(trans.getNode());
				for(Binding binding : sim.getBindings(trans))
				{
					//System.out.println(trans.getNode().getName().asString()+"  "+binding);
					
					int attempt=0;
					
					while(sim.execute(binding) && attempt<10)
					{
						//sim.setMarking(marking);	
						attempt++;
					}
					if (attempt==10)
					{
						//System.err.println("Impossible to execute "+binding+ "at marking:");
						//System.err.println(sim.getMarking());
						continue;
					}
					
					State marking2=sim.getMarking();

					Binding[] path=pathTo.get(marking2);
					if (path==null)
					{
						list.add(marking2);
						Binding[] previousInPath=pathTo.get(marking);
						path=Arrays.copyOf(previousInPath, previousInPath.length+1);
						path[previousInPath.length]=binding;
						pathTo.put(marking2, path);
						if(++numStates % 300==0)
							System.out.println(numStates + " computed... "+" ("+binding+")");
					}

					sim.setMarkingFast(marking,marking2,true);
				}

			}
			if (!enabled)
				deadlocks.add(marking);	
		}
		
		context.log("The reachability graph consists of "+numStates+" states.");
		
		LinkedList<String> placesOfFinalMarking = new LinkedList<String>();
		org.processmining.models.semantics.petrinet.Marking finalMarking = null;
		org.processmining.models.semantics.petrinet.Marking initMarking = null;
		
		try {
			finalMarking = ((org.processmining.models.semantics.petrinet.Marking)context.getConnectionManager()
					.getFirstConnection(FinalMarkingConnection.class, context, dpn)
					.getObjectWithRole(FinalMarkingConnection.MARKING));		
			if (finalMarking.isEmpty()) {
				int yn = JOptionPane.showConfirmDialog(null, "The final marking is empty. Is this intended?",
						"Final Marking", JOptionPane.YES_NO_OPTION);
				if (yn == JOptionPane.NO_OPTION) {
					throw(new IllegalArgumentException());
				}
			}
		} catch (IllegalArgumentException | ConnectionCannotBeObtained e) {
			SortedSet<Place> var = ConvertToCPNTools.createMarking(context, dpn, "Final Marking");
			if (var==null)
			{
				context.getFutureResult(0).cancel(true);
				return null;
			}			
			finalMarking = new org.processmining.models.semantics.petrinet.Marking(var);

			ProvidedObjectHelper.publish
				(context, "Final Marking for " + dpn.getLabel(), 
						finalMarking, org.processmining.models.semantics.petrinet.Marking.class, false);
			context.getConnectionManager().addConnection(new FinalMarkingConnection(dpn, finalMarking));
		}
		for(Place p : finalMarking.baseSet())
			placesOfFinalMarking.add(p.getLabel());		
		
		try {
			initMarking = ((org.processmining.models.semantics.petrinet.Marking)context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, dpn)
					.getObjectWithRole(InitialMarkingConnection.MARKING));		
			if (initMarking.isEmpty()) {
				int yn = JOptionPane.showConfirmDialog(null, "The initial marking is empty. Is this intended?",
						"Initial Marking", JOptionPane.YES_NO_OPTION);
				if (yn == JOptionPane.NO_OPTION) {
					throw(new IllegalArgumentException());
				}
			}
		} catch (IllegalArgumentException | ConnectionCannotBeObtained e) {
			initMarking = new org.processmining.models.semantics.petrinet.Marking
					(ConvertToCPNTools.createMarking(context, dpn, "Initial Marking"));
			ProvidedObjectHelper.publish
				(context, "Initial Marking for " + dpn.getLabel(), 
					initMarking, org.processmining.models.semantics.petrinet.Marking.class, false);
			context.getConnectionManager().addConnection(new InitialMarkingConnection(dpn, initMarking));
		}			
		
		//StringBuffer sb=new StringBuffer();
		CPNAnalysisResults resOutput=new CPNAnalysisResults(dpn,initMarking,finalMarking);
		if (notObservedTransition.size()>0)
		{
		for(Transition t : notObservedTransition)
			resOutput.addDeadTransition(t);
		}
	
		for(State deadlock : deadlocks)
		{
			boolean toBeConsideredAsDeadlock=false;
			for(Marking mark : deadlock.getAllMarkings())
			{
				String placeName=mark.getPlaceInstance().getNode().getName().asString();
				if (!placeName.startsWith("Variable") && !placeName.startsWith("Potential") && !mark.getMarking().trim().equals("empty"))
				{
					for(String placeOfFinalMarkingAsString : placesOfFinalMarking)
					{
						if(!placeName.startsWith(placeOfFinalMarkingAsString))
						{
							toBeConsideredAsDeadlock=true;
							break;
						}
					}
					if (toBeConsideredAsDeadlock)
						break;
				}
				if (toBeConsideredAsDeadlock)
					break;	
			}

			if (toBeConsideredAsDeadlock)
			{
				resOutput.addDeadlock(deadlock,pathTo.get(deadlock));
			}
		}
		
		
		return resOutput;
	}
	
	protected final static HighLevelSimulator checkModel(final PetriNet petriNet, final Progress progress, final Logger logger)
			throws Exception {
		logger.log("Starting CPN Simulator");
		final HighLevelSimulator highLevelSimulator = HighLevelSimulator.getHighLevelSimulator();
		final Checker checker = new Checker(petriNet, null, highLevelSimulator);
		logger.log("Checking CPN model");
		int worked = 0;
		worked = localCheck(worked, checker, progress, logger);
		worked = checkInitializing(worked, checker, progress, logger);
		worked = checkDeclarations(worked, checker, petriNet, progress, logger);
		worked = checkPages(worked, checker, petriNet, progress, logger);
		worked = generateInstances(worked, checker, petriNet, progress, logger);
		checker.initialiseSimulationScheduler();
		try {
			worked = instantiateSMLInterface(worked, checker, progress, logger);
		} catch (final ErrorInitializingSMLInterface e) {
			// Mask this; don't work for timed models
		}
		return highLevelSimulator;
	}	
	
	private static int checkDeclarations(int worked, final Checker checker, final PetriNet petriNet, final Progress progress,
			final Logger logger) throws Exception {
		logger.log("Checking declarations");
		for (final HLDeclaration decl : petriNet.declaration()) {
			checker.checkDeclaration(decl);
			worked++;
			progress.setValue(worked);
		}
		checker.generateSerializers();
		worked++;
		progress.setValue(worked);
		return worked;
	}

	private static int checkInitializing(int worked, final Checker checker, final Progress progress, final Logger logger)
			throws IOException, Exception {
		logger.log("Initializing");
		checker.checkInitializing();
		worked++;
		progress.setValue(worked);
		return worked;
	}
	
	private static int localCheck(int worked, final Checker checker, final Progress progress, final Logger logger)
			throws LocalCheckFailed {
		logger.log("Checking names");
		checker.localCheck();
		worked++;
		progress.setValue(worked);
		return worked;
	}
	
	private static int checkPages(int worked, final Checker checker, final PetriNet petriNet, final Progress progress,
			final Logger logger) throws IOException, CheckerException {
		logger.log("Checking pages");
		final PageSorter pageSorter = new PageSorter(petriNet.getPage());
		for (final Page page : pageSorter) {
			progress.setCaption("Checking page " + page.getName().getText());
			checker.checkPage(page, pageSorter.isPrime(page));
			worked += 3;
			progress.setValue(worked);
		}
		return worked;
	}
	
	private static int generateInstances(int worked, final Checker checker, final PetriNet petriNet, final Progress progress,
			final Logger logger) throws IOException {
		logger.log("Generating instances");

		for (final FusionGroup fusionGroup : petriNet.getFusionGroups()) {
			checker.generateInstanceForFusionGroup(fusionGroup.getId());
			worked++;
			progress.setValue(worked);
		}

		for (final Page page : petriNet.getPage()) {
			logger.log("Generating instances for " + page.getName().getText());

			for (final RefPlace refPlace : page.readyFusionGroups()) {
				checker.generateInstanceForPlace(refPlace.getId());
				worked++;
				progress.setValue(worked);
			}
			for (final RefPlace refPlace : page.readyPortPlaces()) {
				checker.generateInstanceForPlace(refPlace.getId());
				worked++;
				progress.setValue(worked);
			}
			for (final org.cpntools.accesscpn.model.Place place : page.readyPlaces()) {
				checker.generateInstanceForPlace(place.getId());
				worked++;
				progress.setValue(worked);
			}

			for (final Transition transition : page.readyTransitions()) {
				checker.generateInstanceForTransition(transition.getId());
				worked++;
				progress.setValue(worked);
			}
		}

		return worked;
	}

	private static int instantiateSMLInterface(int worked, final Checker checker, final Progress progress, final Logger logger)
			throws ErrorInitializingSMLInterface {
		logger.log("Instantiating SML interface");
		checker.instantiateSMLInterface();
		worked++;
		progress.setValue(worked);
		return worked;
	}	
		
}
