package org.processmining.datapetrinets.visualization.graphviz;

import java.awt.Color;
import java.util.Collections;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.lang3.StringEscapeUtils;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.GuardDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.PlaceDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.VariableDisplayMode;
import org.processmining.models.graphbased.AbstractGraphElement;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.VariableAccess;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableBiMap;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;

public class DPNGraphvizConverter {

	private static final String DEFAULT_FONT = "Myriad Pro,Helvetica";

	public enum DecorationKey {
		EXTRALABEL {
			public void decorate(DotElement node, Object value) {
				final String extraLabel = String.format("%s", (String) value);
				if (!extraLabel.isEmpty()) {
					if (node instanceof DotEdge) {
						node.setLabel("<" + extraLabel + ">");
					} else {
						node.setOption("xlabel", extraLabel); // html-like label not supported by GraphViz package 
					}
				}
			}
		},
		TEXTCOLOR {
			public void decorate(DotElement node, Object value) {
				node.setOption("fontcolor", convertColor((Color) value));
			}
		},
		LINECOLOR {
			public void decorate(DotElement node, Object value) {
				node.setOption("color", String.format("%s", convertColor((Color) value)));
			}
		},
		LINEWIDTH {
			public void decorate(DotElement node, Object value) {
				node.setOption("penwidth", String.format("%.2f", ((Float) value)));
			}
		},
		FILLCOLOR {
			public void decorate(DotElement node, Object value) {
				node.setOption("fillcolor", String.format("%s", convertColor((Color) value)));
				node.setOption("style", "filled");
			}
		},
		LINESTYLE {
			public void decorate(DotElement node, Object value) {
				node.setOption("style", "dotted");
			}
		};

		public abstract void decorate(DotElement node, Object value);

	}

	public interface DPNAsDot {
		BiMap<PetrinetNode, DotNode> getMapping();

		Map<PetrinetEdge<?, ?>, DotEdge> getEdgeMapping(); // one DotEdge might be mapped to two PetrinetEdges in case of bi-directional edges 

		Dot getDot();
	}

	public static DPNAsDot convertDPN(DataPetriNet dpn) {
		return convertDPN(dpn, new Marking(), new Marking[] { new Marking() }, GuardDisplayMode.EDGES,
				VariableDisplayMode.HIDDEN, PlaceDisplayMode.BASIC, GraphDirection.leftRight,
				Collections.<AbstractGraphElement, Map<DecorationKey, Object>>emptyMap());
	}

	public static DPNAsDot convertDPN(DataPetriNet dpn, Marking initialMarking) {
		return convertDPN(dpn, initialMarking, new Marking[] { new Marking() }, GuardDisplayMode.EDGES,
				VariableDisplayMode.HIDDEN, PlaceDisplayMode.BASIC, GraphDirection.leftRight,
				Collections.<AbstractGraphElement, Map<DecorationKey, Object>>emptyMap());
	}

	public static DPNAsDot convertDPN(DataPetriNet dpn, Marking initialMarking, Marking[] finalMarkings) {
		return convertDPN(dpn, initialMarking, finalMarkings, GuardDisplayMode.EDGES, VariableDisplayMode.HIDDEN,
				PlaceDisplayMode.BASIC, GraphDirection.leftRight,
				Collections.<AbstractGraphElement, Map<DecorationKey, Object>>emptyMap());
	}

	public static DPNAsDot convertDPN(DataPetriNet dpn, Marking initialMarking, Marking[] finalMarkings,
			GuardDisplayMode guardDisplayMode, VariableDisplayMode variableDisplayMode,
			PlaceDisplayMode placeDisplayMode, GraphDirection direction) {
		return convertDPN(dpn, initialMarking, finalMarkings, guardDisplayMode, variableDisplayMode, placeDisplayMode,
				direction, Collections.<AbstractGraphElement, Map<DecorationKey, Object>>emptyMap());
	}

	public static DPNAsDot convertDPN(DataPetriNet dpn, Marking initialMarking, Marking[] finalMarkings,
			final GuardDisplayMode guardDisplayMode, VariableDisplayMode variableMode,
			PlaceDisplayMode placeDisplayMode, GraphDirection direction,
			Map<AbstractGraphElement, Map<DecorationKey, Object>> attributeMap) {
		return new DPNGraphvizConverter().doConvertDPN(dpn, initialMarking, finalMarkings, guardDisplayMode,
				variableMode, placeDisplayMode, direction, attributeMap);
	}

	public DPNAsDot doConvertDPN(DataPetriNet dpn, Marking initialMarking, Marking[] finalMarkings,
			GuardDisplayMode guardDisplayMode, VariableDisplayMode variableMode, PlaceDisplayMode placeDisplayMode,
			GraphDirection direction, Map<AbstractGraphElement, Map<DecorationKey, Object>> attributeMap) {
		final Dot dot = new Dot();
		initDot(direction, dot);

		final Map<PetrinetNode, DotNode> mapPetrinet2Dot = Maps.newHashMapWithExpectedSize(dpn.getNodes().size());

		for (Place place : dpn.getPlaces()) {
			PlaceNode placeNode = convertPlace(initialMarking, finalMarkings, place, placeDisplayMode,
					attributeMap.get(place));
			mapPetrinet2Dot.put(place, placeNode);
			dot.addNode(placeNode);
		}

		for (Transition t : dpn.getTransitions()) {
			TransitionNode transitionNode = convertTransition(t, attributeMap.get(t), guardDisplayMode, variableMode);
			transitionNode.setSelectable(true);
			mapPetrinet2Dot.put(t, transitionNode);
			dot.addNode(transitionNode);
		}

		Map<DataElement, VariableNode> mapVariable2Dot = Maps.newHashMapWithExpectedSize(dpn.getVariables().size());
		switch (variableMode) {
			case NO_LAYOUT :
			case NO_LAYOUT_NO_READS:
				for (DataElement writeOp : dpn.getVariables()) {
					VariableNode node = convertVariable(writeOp, attributeMap.get(writeOp));
					mapVariable2Dot.put(writeOp, node);
					dot.addNode(node);
				}
				setVariablesNoLayout(dpn, dot, mapPetrinet2Dot, mapVariable2Dot, variableMode == variableMode.NO_LAYOUT);
				break;

			case AUTO_LAYOUT :
			case AUTO_LAYOUT_NO_READS:
				for (DataElement writeOp : dpn.getVariables()) {
					VariableNode node = convertVariable(writeOp, attributeMap.get(writeOp));
					mapVariable2Dot.put(writeOp, node);
					dot.addNode(node);
				}

				setVariablesAutoLayout(dpn, dot, mapPetrinet2Dot, mapVariable2Dot, variableMode == variableMode.AUTO_LAYOUT);
				break;

			case HIDDEN :
			default :
				break;

		}

		BiMap<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> biEdgesToSkip = determineSkippedBiEdges(
				dpn);
		final Map<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>, DotEdge> mapEdgeToDot = new HashMap<>();

		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : dpn.getEdges()) {
			if (!(edge instanceof VariableAccess)) {
				DotNode target = mapPetrinet2Dot.get(edge.getTarget());
				DotNode source = mapPetrinet2Dot.get(edge.getSource());
				connectEdge(initialMarking, guardDisplayMode, attributeMap, dot, biEdgesToSkip, mapEdgeToDot, edge,
						target, source);
			}
		}

		final ImmutableBiMap<PetrinetNode, DotNode> mapping = ImmutableBiMap.<PetrinetNode, DotNode>builder()
				.putAll(mapPetrinet2Dot).putAll(mapVariable2Dot).build();

		return new DPNAsDot() {

			public BiMap<PetrinetNode, DotNode> getMapping() {
				return mapping;
			}

			public Map<PetrinetEdge<?, ?>, DotEdge> getEdgeMapping() {
				return mapEdgeToDot;
			}

			public Dot getDot() {
				return dot;
			}

		};
	}

	private static void initDot(GraphDirection direction, final Dot dot) {
		dot.setDirection(direction);
		dot.setKeepOrderingOfChildren(false);
		dot.setOption("ratio", String.format(Locale.ENGLISH, "%.2f", RatioAwareDotPanel.STANDARD_HORIZONTAL_RATIO));
		dot.setOption("nodesep", "0.2");
		dot.setOption("ranksep", "0.2");
		dot.setNodeOption("fontname", DEFAULT_FONT);
		dot.setEdgeOption("fontname", DEFAULT_FONT);
	}

	private void connectEdge(Marking initialMarking, final GuardDisplayMode guardDisplayMode,
			Map<AbstractGraphElement, Map<DecorationKey, Object>> attributeMap, final Dot dot,
			BiMap<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> biEdgesToSkip,
			final Map<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>, DotEdge> mapEdgeToDot,
			PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge, DotNode target, DotNode source) {
		if (target != null && source != null && !biEdgesToSkip.containsKey(edge)) {
			PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> biEdge = biEdgesToSkip.inverse().get(edge);
			DotEdge dotEdge = addEdge(dot, edge, target, source, guardDisplayMode, biEdge != null,
					attributeMap.get(edge), initialMarking);
			mapEdgeToDot.put(edge, dotEdge);
			if (biEdge != null) {
				mapEdgeToDot.put(biEdge, dotEdge);
			}
		}
	}

	private static void setVariablesAutoLayout(DataPetriNet dpn, final Dot dot,
			final Map<PetrinetNode, DotNode> mapPetrinet2Dot, Map<DataElement, VariableNode> mapVariable2Dot, boolean showReads) {
		for (Transition t : dpn.getTransitions()) {
			if (t instanceof PNWDTransition) {
				final PNWDTransition tD = (PNWDTransition) t;
				for (DataElement writeOp : tD.getWriteOperations()) {
					dot.addEdge(mapPetrinet2Dot.get(tD), mapVariable2Dot.get(writeOp), "",
							ImmutableMap.of("style", "dashed", "penwidth", "1"));
				}
				if (showReads) {
					for (DataElement readOp : tD.getReadOperations()) {
						dot.addEdge(mapVariable2Dot.get(readOp), mapPetrinet2Dot.get(tD), "",
								ImmutableMap.of("style", "dashed", "penwidth", "1"));
					}
				}
			}
		}
	}

	private static void setVariablesNoLayout(DataPetriNet dpn, final Dot dot,
			final Map<PetrinetNode, DotNode> mapPetrinet2Dot, Map<DataElement, VariableNode> mapVariable2Dot, boolean showReads) {
		for (Transition t : dpn.getTransitions()) {
			if (t instanceof PNWDTransition) {
				final PNWDTransition tD = (PNWDTransition) t;
				for (DataElement writeOp : tD.getWriteOperations()) {
					dot.addEdge(mapPetrinet2Dot.get(tD), mapVariable2Dot.get(writeOp), "",
							ImmutableMap.of("constraint", "false", "style", "dotted", "penwidth", "2"));
				}
				if (showReads) {
					for (DataElement readOp : tD.getReadOperations()) {
						dot.addEdge(mapVariable2Dot.get(readOp), mapPetrinet2Dot.get(tD), "",
								ImmutableMap.of("constraint", "false", "style", "dotted", "penwidth", "2"));
					}
				}
			}
		}
	}

	/**
	 * @param dpn
	 * @return a map from to be skipped edges to the not-skipped edges in the
	 *         other direction
	 */
	private static BiMap<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> determineSkippedBiEdges(
			DataPetriNet dpn) {
		BiMap<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> biEdgesToSkip = HashBiMap
				.create();

		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : dpn.getEdges()) {
			if (edge.getSource() instanceof Place) {
				Place sourcePlace = (Place) edge.getSource();
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdgeForPlace : dpn
						.getInEdges(sourcePlace)) {
					if (inEdgeForPlace.getSource().equals(edge.getTarget())) {
						biEdgesToSkip.put(inEdgeForPlace, edge);
					}
				}
			}
		}
		return biEdgesToSkip;
	}

	private int maxGuardLabelLength = 101;
	private int maxGuardLabelLineLength = 30;

	private int maxLabelLength = 27;
	private int maxLabelLineLength = 8;

	private DotEdge addEdge(final Dot dot, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge,
			DotNode target, DotNode source, GuardDisplayMode guardDisplayMode, boolean isBiEdge,
			Map<DecorationKey, Object> attributeMap, Marking initialMarking) {
		DotEdge dotEdge = null;
		if (edge.getTarget() instanceof PNWDTransition) {
			PNWDTransition t = (PNWDTransition) edge.getTarget();
			if (t.getGuardExpression() != null && !guardAlwaysTrue(t) && guardDisplayMode == GuardDisplayMode.EDGES) {
				dotEdge = dot.addEdge(source, target, convertGuardToLabel(t.getGuardExpression(),
						getMaxGuardLabelLength(), getMaxGuardLabelLineLength()));
			} else {
				dotEdge = dot.addEdge(source, target, "");
			}
		} else {
			dotEdge = dot.addEdge(source, target, "");
		}

		if (edge instanceof Arc) {
			int weight = ((Arc) edge).getWeight();
			if (weight > 1) {
				if (dotEdge.getLabel().equals("")) {
					dotEdge.setLabel(String.valueOf(weight));
				} else {
					dotEdge.setOption("xlabel", String.valueOf(weight));
				}
			}
		}

		dotEdge.setOption("style", "bold");
		if (isBiEdge) {
			dotEdge.setOption("dir", "both");
		}
		if (initialMarking != null && initialMarking.contains(edge.getTarget())) {
			dotEdge.setOption("constraint", "false");
		}
		if (attributeMap != null) {
			decorateNode(attributeMap, dotEdge);
		}
		return dotEdge;
	}

	private static boolean guardAlwaysTrue(PNWDTransition t) {
		return t.getGuardExpression().toCanonicalString().equals("true");
	}

	private static String convertGuardToLabel(GuardExpression guard, int maxOverallLength, int maxLineLength) {
		String label = guard.toPrettyString(1);
		if (label.length() > maxOverallLength) {
			label = org.apache.commons.lang3.StringUtils.abbreviate(label, maxOverallLength);
		}
		label = StringEscapeUtils.escapeXml10(label);
		if (label.length() > maxLineLength) {
			label = VisualizationUtils.wrapLineWithSeparator(label, "<BR/>", maxLineLength);
		}
		return "<" + label + ">";
	}

	private String convertTransitionLabel(String label) {
		if (label.isEmpty()) {
			return label;
		} else {
			String escapeXml = StringEscapeUtils.escapeXml10(label);
			if (escapeXml.length() > getMaxLabelLength()) {
				escapeXml = org.apache.commons.lang3.StringUtils.abbreviate(escapeXml, getMaxLabelLength());
			}
			if (escapeXml.length() > getMaxLabelLineLength()) {
				escapeXml = VisualizationUtils.wrapLineWithSeparator(escapeXml, "<BR/>", getMaxLabelLineLength());
			}
			return "<<B>" + escapeXml + "</B>>";
		}
	}

	private static PlaceNode convertPlace(Marking initialMarking, Marking[] finalMarkings, Place place,
			PlaceDisplayMode placeDisplayMode, Map<DecorationKey, Object> attributeMap) {
		PlaceNode placeNode = null;
		switch (placeDisplayMode) {
			case BASIC :
				placeNode = new PlaceNode();
				break;
			case DETAILED :
				placeNode = new PlaceNode(place.getLabel());
				break;
		}
		if (initialMarking != null && initialMarking.contains(place) && finalMarkings != null
				&& markingsContains(finalMarkings, place)) {
			Integer numTokens = initialMarking.occurrences(place);
			if (numTokens > 1) {
				placeNode.setLabel(String.valueOf(numTokens));
			} else {
				placeNode.setLabel("\u2022");
			}
			placeNode.setOption("style", "bold");
			placeNode.setOption("penwidth", "4");
		} else if (initialMarking != null && initialMarking.contains(place)) {
			Integer numTokens = initialMarking.occurrences(place);
			if (numTokens > 1) {
				placeNode.setLabel(String.valueOf(numTokens));
			} else {
				placeNode.setLabel("\u2022");
			}
		} else if (finalMarkings != null && markingsContains(finalMarkings, place)) {
			placeNode.setOption("style", "bold");
			placeNode.setOption("penwidth", "4");
			Integer numTokens = finalMarkings[0].occurrences(place);
			if (numTokens > 1) {
				placeNode.setLabel(String.valueOf(numTokens));
			} else {
				placeNode.setLabel("F");
			}
		}
		if (attributeMap != null) {
			decorateNode(attributeMap, placeNode);
		}
		placeNode.setSelectable(true);
		return placeNode;
	}

	private static String convertColor(Color color) {
		return String.format("#%02X%02X%02X", color.getRed(), color.getGreen(), color.getBlue());
	}

	private static boolean markingsContains(Marking[] finalMarkings, Place place) {
		for (Marking m : finalMarkings) {
			if (m.contains(place)) {
				return true;
			}
		}
		return false;
	}

	private TransitionNode convertTransition(final Transition t, Map<DecorationKey, Object> attributeMap,
			GuardDisplayMode guardDisplayMode, VariableDisplayMode variableMode) {
		final TransitionNode transitionNode;

		if (t instanceof PNWDTransition && ((PNWDTransition) t).getGuardExpression() != null) {
			GuardExpression guardExpression = ((PNWDTransition) t).getGuardExpression();
			if (t.isInvisible()) {
				transitionNode = new TransitionNode(t, guardExpression);
			} else {
				transitionNode = new TransitionNode(t, convertTransitionLabel(t.getLabel()), guardExpression);
			}
			if (guardDisplayMode == GuardDisplayMode.TRANSITIONS) {
				transitionNode.setOption("xlabel", String.format("%s", convertGuardToLabel(guardExpression,
						getMaxGuardLabelLineLength(), getMaxGuardLabelLineLength() * 2)));
			}
		} else {
			if (t.isInvisible()) {
				transitionNode = new TransitionNode(t);
			} else {
				transitionNode = new TransitionNode(t, convertTransitionLabel(t.getLabel()));
			}
		}

		if (attributeMap != null) {
			decorateNode(attributeMap, transitionNode);
		}
		transitionNode.setSelectable(true);
		return transitionNode;
	}

	public static void decorateNode(Map<DecorationKey, Object> decorationMap, final DotElement node) {
		for (Entry<DecorationKey, Object> entry : decorationMap.entrySet()) {
			entry.getKey().decorate(node, entry.getValue());
		}
	}

	private static VariableNode convertVariable(DataElement writeOp, Map<DecorationKey, Object> attributeMap) {
		VariableNode variableNode = new VariableNode(writeOp.getVarName());
		if (attributeMap != null) {
			decorateNode(attributeMap, variableNode);
		}
		variableNode.setSelectable(true);
		return variableNode;
	}

	public int getMaxLabelLength() {
		return maxLabelLength;
	}

	public void setMaxLabelLength(int maxLabelLength) {
		this.maxLabelLength = maxLabelLength;
	}

	public int getMaxLabelLineLength() {
		return maxLabelLineLength;
	}

	public void setMaxLabelLineLength(int maxLabelLineLength) {
		this.maxLabelLineLength = maxLabelLineLength;
	}

	public int getMaxGuardLabelLength() {
		return maxGuardLabelLength;
	}

	public void setMaxGuardLabelLength(int maxGuardLabelLength) {
		this.maxGuardLabelLength = maxGuardLabelLength;
	}

	public int getMaxGuardLabelLineLength() {
		return maxGuardLabelLineLength;
	}

	public void setMaxGuardLabelLineLength(int maxGuardLabelLineLength) {
		this.maxGuardLabelLineLength = maxGuardLabelLineLength;
	}

}
