package org.processmining.datapetrinets.visualization.graphviz;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.geom.NoninvertibleTransformException;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.graphviz.visualisation.listeners.SelectionChangedListener;

import com.kitfox.svg.SVGDiagram;

public class RatioAwareDotPanel extends JPanel {

	private final class DotPanelRatioAware extends DotPanel {

		private static final long serialVersionUID = 1L;

		private DotPanelRatioAware(Dot dot) {
			super(dot);
			helperControlsShortcuts.add("ctrl r");
			helperControlsExplanations.add("show aspect ratio slider");
			getInputMap(WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_R, InputEvent.CTRL_MASK),
					"showAspectRatioSlider");
			getActionMap().put("showAspectRatioSlider", new AbstractAction("showAspectRatioSlider") {

				private static final long serialVersionUID = 1L;

				public void actionPerformed(ActionEvent e) {
					ratioSlider.setVisible(!ratioSlider.isVisible());
				}
			});
		}

		public void changeDot(Dot dot, boolean resetView) {
			GraphDirection direction = getDotPanel().getUserSettings().direction;
			setupRatioDefault(direction);
			dot.setOption("ratio", genRatio(ratioSlider));
			super.changeDot(dot, resetView);
		}

	}

	private static final long serialVersionUID = 1529723568420594808L;

	public static final float STANDARD_HORIZONTAL_RATIO = 0.3f;
	public static final float STANDARD_VERTICAL_RATIO = 1.3f;

	private final JSlider ratioSlider;
	private final DotPanel dotPanel;

	public RatioAwareDotPanel(Dot dot) {
		super();
		setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
		setBackground(Color.WHITE);
		setOpaque(true);
		dotPanel = new DotPanelRatioAware(dot);

		ratioSlider = new JSlider();
		ratioSlider.setToolTipText("Changes the aspect ratio of the displayed graph");
		ratioSlider.setBackground(Color.WHITE);
		ratioSlider.setMaximumSize(new Dimension(40, 200));
		ratioSlider.setMinimum(1);
		ratioSlider.setMaximum(2000);
		ratioSlider.setValue(genSliderValue(STANDARD_HORIZONTAL_RATIO));
		ratioSlider.setOrientation(SwingConstants.VERTICAL);
		ratioSlider.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider) e.getSource();
				if (!source.getValueIsAdjusting()) {
					getDotPanel().getDot().setOption("ratio", genRatio(source));
					changeDot(getDotPanel().getDot(), false);
				}
			}

		});
		ratioSlider.setVisible(false);
		add(ratioSlider);
		add(dotPanel);
	}

	private static int genSliderValue(double ratio) {
		return (int) (ratio * 1000);
	}

	private void setupRatioDefault(GraphDirection direction) {
		if (isDefault(ratioSlider.getValue())) {
			if (isHorizontal(direction)) {
				ratioSlider.setValue(genSliderValue(STANDARD_HORIZONTAL_RATIO));
			} else {
				ratioSlider.setValue(genSliderValue(STANDARD_VERTICAL_RATIO));
			}
		}
	}

	private boolean isDefault(int value) {
		return value == genSliderValue(STANDARD_HORIZONTAL_RATIO) || value == genSliderValue(STANDARD_VERTICAL_RATIO);
	}

	private boolean isHorizontal(GraphDirection direction) {
		return direction == GraphDirection.leftRight || direction == GraphDirection.rightLeft;
	}

	public SVGDiagram generateSVG(Dot dot) {
		GraphDirection direction = getDotPanel().getUserSettings().direction;
		setupRatioDefault(direction); // query dot panel for direction
		dot.setOption("ratio", genRatio(ratioSlider));
		dot.setDirection(direction);
		return DotPanel.dot2svg(dot);
	}

	public void changeDot(Dot dot, boolean resetView) {
		getDotPanel().changeDot(dot, resetView);
	}

	public void changeDot(Dot dot, SVGDiagram diagram, boolean resetView) {
		getDotPanel().changeDot(dot, diagram, resetView);
	}

	public void resetView() throws NoninvertibleTransformException {
		if (dotPanel.isVisible() && dotPanel.getWidth() > 0 && dotPanel.getHeight() > 0) {
			dotPanel.resetView();
		}
	}

	private static String genRatio(JSlider source) {
		return String.format(Locale.ENGLISH, "%.2f", ratio(source));
	}

	private static float ratio(JSlider source) {
		return source.getValue() / 1000.0f;
	}

	public DotPanel getDotPanel() {
		return dotPanel;
	}

	public Dot getDot() {
		return dotPanel.getDot();
	}

	public List<String> getHelperControlsExplanations() {
		return dotPanel.getHelperControlsExplanations();
	}

	public List<String> getHelperControlsShortcuts() {
		return dotPanel.getHelperControlsShortcuts();
	}

	public Set<DotElement> getSelectedElements() {
		return dotPanel.getSelectedElements();
	}

	public void addSelectionChangedListener(SelectionChangedListener<DotElement> listener) {
		dotPanel.addSelectionChangedListener(listener);
	}

	public SVGDiagram getSVG() {
		return dotPanel.getSVG();
	}

	public void select(DotNode element) {
		dotPanel.select(element);
	}

}
