package org.processmining.models.graphbased.directed.petrinetwithdata.newImpl;

import java.awt.Color;
import java.io.Serializable;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;

import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.AbstractDirectedGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.guards.Expression;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;

public class PNWDTransition extends Transition implements Serializable {

	private static final long serialVersionUID = -6726503006988773331L;

	private final Set<DataElement> readOperations = new HashSet<>();
	private final Set<DataElement> writeOperations = new HashSet<>();

	/**
	 * A rule that defines whether this transition is enabled or not
	 */
	private GuardExpression guard = null;

	/**
	 * Some quality measure in the range [0..1] that expressed the quality of
	 * the guard (if it was discovered).
	 */
	private double quality = -1.0d;

	public PNWDTransition(String label,
			AbstractDirectedGraph<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> net,
			ExpandableSubNet parent) {
		super(label, net, parent);
		updateLabel();
	}

	public PNWDTransition(String label,
			AbstractDirectedGraph<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> net) {
		this(label, net, null);
	}

	private void updateLabel() {
		if (guard != null) {
			getAttributeMap().put(AttributeMap.LABELCOLOR, Color.RED);
			if (this.isInvisible())
				getAttributeMap().put(AttributeMap.FILLCOLOR, Color.RED);
		} else {
			getAttributeMap().put(AttributeMap.LABELCOLOR, Color.BLACK);
			if (this.isInvisible())
				getAttributeMap().put(AttributeMap.FILLCOLOR, Color.BLACK);
		}
		this.getGraph().graphElementChanged(this);
	}

	void assignReadOperation(DataPetriNet petriNetWithData, Transition t, DataElement v) {
		if (this.getGraph() != petriNetWithData)
			throw (new IllegalArgumentException("The transition belongs to a different net"));
		readOperations.add(v);
	}

	void removeReadOperation(DataPetriNet petriNetWithData, Transition t, DataElement v) {
		if (this.getGraph() != petriNetWithData)
			throw (new IllegalArgumentException("The transition belongs to a different net"));
		readOperations.remove(v);
	}

	void assignWriteOperation(DataPetriNet petriNetWithData, Transition t, DataElement v) {
		if (this.getGraph() != petriNetWithData)
			throw (new IllegalArgumentException("The transition belongs to a different net"));
		writeOperations.add(v);
	}

	void removeWriteOperation(DataPetriNet petriNetWithData, Transition t, DataElement v) {
		if (this.getGraph() != petriNetWithData)
			throw (new IllegalArgumentException("The transition belongs to a different net"));
		writeOperations.remove(v);
	}

	public void removeGuard() {
		this.guard = null;
		updateLabel();
	}

	public Set<DataElement> getReadOperations() {
		return readOperations;
	}

	public Set<DataElement> getWriteOperations() {
		return writeOperations;
	}

	public String getGuardAsString() {
		if (guard != null) {
			return guard.toCanonicalString();
		} else {
			return null;
		}
	}

	public void setGuard(DataPetriNet petriNetWithData, String expression)
			throws NonExistingVariableException, ParseException {
		if (expression == null) {
			this.guard = null;
		}
		else
			setGuard(petriNetWithData, GuardExpression.Factory.newInstance(expression));
	}

	/**
	 * Associates a {@link GuardExpression} with this transition.
	 * 
	 * @param petriNetWithData
	 * @param guardExpression
	 * @throws NonExistingVariableException
	 *             in case the expression refers to a missing variable
	 */
	public void setGuard(DataPetriNet petriNetWithData, GuardExpression guardExpression)
			throws NonExistingVariableException {
		if (guardExpression == null) {
			this.guard = null;
		}
		else
		{
			if (this.getGraph() != petriNetWithData)
				throw (new IllegalArgumentException("The transition belongs to a different net"));
			checkNonExistingVariables(petriNetWithData.getVariables(), guardExpression);
			guard = guardExpression;
		}
	}
	
	/**
	 * Associates a {@link GuardExpression} with this transition.
	 * 
	 * @param petriNetWithData
	 * @param guardExpression
	 * @param quality
	 * @throws NonExistingVariableException
	 *             in case the expression refers to a missing variable
	 */
	public void setGuard(DataPetriNet petriNetWithData, GuardExpression guardExpression, double quality)
			throws NonExistingVariableException {
		setGuard(petriNetWithData, guardExpression);
		this.quality = quality;
	}

	/**
	 * @return the guard expression associated with this transition
	 */
	public GuardExpression getGuardExpression() {
		return guard;
	}
	
	/**
	 * @return whether a guard expression is associated with this transition
	 */
	public boolean hasGuardExpression() {
		return guard != null;
	}

	private static void checkNonExistingVariables(Collection<DataElement> dpnVariables, GuardExpression guardExpression)
			throws NonExistingVariableException {
		Set<String> normalVariables = guardExpression.getNormalVariables();
		Set<String> primeVariables = guardExpression.getPrimeVariables();
		Set<String> dpnVariableNames = new HashSet<>();
		for (DataElement element : dpnVariables) {
			dpnVariableNames.add(element.getVarName());
		}
		ImmutableSet<String> missingVariables = ImmutableSet.copyOf(Sets.difference(Sets.union(normalVariables, primeVariables), dpnVariableNames));
		if (!missingVariables.isEmpty()) {
			throw new NonExistingVariableException(missingVariables);
		}
	}

	/**
	 * @return a value from [0..1], or -1 in case there is not quality measure
	 *         available
	 */
	public double getQuality() {
		return quality;
	}

	/*
	 * Old depreciated methods only retained until all packages are updated to
	 * the new format
	 */
	
	/**
	 * 
	 */
	@Deprecated
	public void setGuardAsString(DataPetriNet petriNetWithData, String guardAsString) throws java.text.ParseException {
		if (this.getGraph() != petriNetWithData)
			throw (new IllegalArgumentException("The transition belongs to a different net"));
		try {
			setGuard(petriNetWithData, GuardExpression.Factory.newInstance(guardAsString));
		} catch (org.processmining.datapetrinets.expression.syntax.ParseException e) {
			throw new java.text.ParseException(e.getMessage(), -1);
		} catch (NonExistingVariableException e) {
			throw new IllegalArgumentException(e);
		}
	}

	/**
	 * Use {@link #setGuard(PetriNetWithData, GuardExpression)} instead
	 */
	@Deprecated
	public void setGuard(DataPetriNet petriNetWithData, Expression expr) {
		if (this.getGraph() != petriNetWithData)
			throw new IllegalArgumentException("The transition belongs to a different net");
		String varName = containNonExistingVariable(petriNetWithData, expr);
		if (varName != null)
			throw new IllegalArgumentException("The expression is defined over an undefined variable " + varName);
		try {
			this.guard = GuardExpression.Factory.newInstance(expr.toString());
		} catch (org.processmining.datapetrinets.expression.syntax.ParseException e) {
			throw new RuntimeException("Could not convert old Expression into a GuardExpression!", e);
		}
		updateLabel();
	}

	@Deprecated
	private String containNonExistingVariable(DataPetriNet petriNetWithData, Expression operand) {
		@SuppressWarnings("rawtypes")
		Enumeration variableEnumeration = operand.findVariables();
		while (variableEnumeration.hasMoreElements()) {
			String variableInExpression = ((String) variableEnumeration.nextElement());
			if (petriNetWithData.getVariable(variableInExpression.replace(PetriNetWithData.WRITE_CHAR, ' ').trim()) == null)
				return variableInExpression;
		}
		return null;
	}

	/**
	 * Use {@link #getGuardExpression()} instead
	 */
	@Deprecated
	public Expression getGuard() {
		try {
			if (guard == null) {
				return null;
			}
			// A bridge between the old and new world
			return new org.processmining.models.guards.Expression(guard.toCanonicalString());
		} catch (java.text.ParseException e) {
			throw new IllegalArgumentException("Could not transform the guard into the old representation!", e);
		}
	}

	/**
	 * Use {@link #setGuard(PetriNetWithData, GuardExpression)} instead
	 */
	@Deprecated
	public void setGuard(DataPetriNet petriNetWithData, Expression guard2, double quality) {
		this.quality = quality;
		this.setGuard(petriNetWithData, guard2);
	}

	@Deprecated
	public double getConfidence() {
		return quality;
	}
	
}
